@@ -91,7 +91,12 @@ use {
     bcder::{Integer, OctetString, Oid},
     pem::PemError,
     ring::{digest::Digest, signature::UnparsedPublicKey},
-    std::{collections::HashSet, convert::TryFrom, fmt::Display, ops::Deref},
+    std::{
+        collections::HashSet,
+        convert::TryFrom,
+        fmt::{Debug, Display, Formatter},
+        ops::Deref,
+    },
     x509_certificate::{
         certificate::certificate_is_subset_of, rfc3280::Name, CapturedX509Certificate,
         DigestAlgorithm, SignatureAlgorithm, X509Certificate, X509CertificateError,
@@ -173,9 +178,9 @@ pub enum CmsError {
 impl std::error::Error for CmsError {}
 
 impl Display for CmsError {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
         match self {
-            Self::DecodeErr(e) => e.fmt(f),
+            Self::DecodeErr(e) => std::fmt::Display::fmt(e, f),
             Self::MissingSignedAttributeContentType => {
                 f.write_str("content-type attribute missing from SignedAttributes")
             }
@@ -197,7 +202,7 @@ impl Display for CmsError {
             Self::SubjectKeyIdentifierUnsupported => {
                 f.write_str("signer info using subject key identifier is not supported")
             }
-            Self::Io(e) => e.fmt(f),
+            Self::Io(e) => std::fmt::Display::fmt(e, f),
             Self::UnknownKeyAlgorithm(oid) => {
                 f.write_fmt(format_args!("unknown signing key algorithm: {}", oid))
             }
@@ -268,7 +273,7 @@ impl From<X509CertificateError> for CmsError {
 /// This is a high-level data structure that ultimately gets (de)serialized
 /// from/to ASN.1. It exists to facilitate common interactions with the
 /// low-level ASN.1 without exposing the complexity of ASN.1.
-#[derive(Clone, Debug)]
+#[derive(Clone)]
 pub struct SignedData {
     /// Content digest algorithms used.
     digest_algorithms: HashSet<DigestAlgorithm>,
@@ -293,6 +298,20 @@ pub struct SignedData {
     signers: Vec<SignerInfo>,
 }
 
+impl Debug for SignedData {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        let mut s = f.debug_struct("SignedData");
+        s.field("digest_algorithms", &self.digest_algorithms);
+        s.field(
+            "signed_content",
+            &format_args!("{:?}", self.signed_content.as_ref().map(hex::encode)),
+        );
+        s.field("certificates", &self.certificates);
+        s.field("signers", &self.signers);
+        s.finish()
+    }
+}
+
 impl SignedData {
     /// Construct an instance by parsing BER data.
     pub fn parse_ber(data: &[u8]) -> Result<Self, CmsError> {
@@ -352,11 +371,11 @@ impl TryFrom<&crate::asn1::rfc5652::SignedData> for SignedData {
             .map(DigestAlgorithm::try_from)
             .collect::<Result<HashSet<_>, _>>()?;
 
-        let signed_content = if let Some(content) = &raw.content_info.content {
-            Some(content.to_bytes().to_vec())
-        } else {
-            None
-        };
+        let signed_content = raw
+            .content_info
+            .content
+            .as_ref()
+            .map(|content| content.to_bytes().to_vec());
 
         let certificates = if let Some(certs) = &raw.certificates {
             Some(
@@ -402,7 +421,7 @@ impl TryFrom<&crate::asn1::rfc5652::SignedData> for SignedData {
 /// Instances of this type are logically equivalent to a single
 /// signed assertion within a `SignedData` payload. There can be multiple
 /// signers per `SignedData`, which is why this type exists on its own.
-#[derive(Clone, Debug)]
+#[derive(Clone)]
 pub struct SignerInfo {
     /// The X.509 certificate issuer.
     issuer: Name,
@@ -429,6 +448,32 @@ pub struct SignerInfo {
     unsigned_attributes: Option<UnsignedAttributes>,
 }
 
+impl Debug for SignerInfo {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        let mut s = f.debug_struct("SignerInfo");
+        s.field("issuer", &self.issuer);
+        s.field("serial_number", &self.serial_number);
+        s.field("digest_algorithm", &self.digest_algorithm);
+        s.field("signature_algorithm", &self.signature_algorithm);
+        s.field(
+            "signature",
+            &format_args!("{}", hex::encode(&self.signature)),
+        );
+        s.field("signed_attributes", &self.signed_attributes);
+        s.field(
+            "digested_signed_attributes_data",
+            &format_args!(
+                "{:?}",
+                self.digested_signed_attributes_data
+                    .as_ref()
+                    .map(hex::encode)
+            ),
+        );
+        s.field("unsigned_attributes", &self.unsigned_attributes);
+        s.finish()
+    }
+}
+
 impl SignerInfo {
     /// Obtain the signing X.509 certificate's issuer name and its serial number.
     ///
@@ -480,6 +525,7 @@ impl SignerInfo {
     /// * DOES NOT verify the digest hash embedded within `SignedAttributes` (if present).
     /// * DOES NOT validate the signing certificate in any way.
     /// * DOES NOT validate that the cryptography used is appropriate.
+    /// * DOES NOT verify the time stamp token, if present.
     ///
     /// See the crate's documentation for more on the security implications.
     pub fn verify_signature_with_signed_data(
@@ -538,6 +584,35 @@ impl SignerInfo {
         }
     }
 
+    /// Verifies the migest digest stored in signed attributes using explicit encapsulated content.
+    ///
+    /// Typically, the digest is computed over content stored in the [SignedData] instance.
+    /// However, it is possible for the signed content to be external. This function
+    /// allows you to define the source of that external content.
+    ///
+    /// Behavior is very similar to [SignerInfo::verify_message_digest_with_signed_data]
+    /// except the original content that was digested is explicitly passed in. This
+    /// content is appended with the signed attributes data on this [SignerInfo].
+    ///
+    /// The security limitations from [SignerInfo::verify_message_digest_with_signed_data]
+    /// apply to this function as well.
+    pub fn verify_message_digest_with_content(&self, data: &[u8]) -> Result<(), CmsError> {
+        let signed_attributes = self
+            .signed_attributes()
+            .ok_or(CmsError::NoSignedAttributes)?;
+
+        let wanted_digest: &[u8] = signed_attributes.message_digest.as_ref();
+        let got_digest = self.compute_digest(Some(data));
+
+        // Susceptible to timing side-channel but we don't care per function
+        // documentation.
+        if wanted_digest == got_digest.as_ref() {
+            Ok(())
+        } else {
+            Err(CmsError::DigestNotEqual)
+        }
+    }
+
     /// Obtain an entity for validating the signature described by this instance.
     ///
     /// This will attempt to locate the certificate used by this signing info
@@ -577,12 +652,9 @@ impl SignerInfo {
             CmsError::UnknownKeyAlgorithm(signing_cert.key_algorithm_oid().clone())
         })?;
 
-        let signature_algorithm = signing_cert.signature_algorithm().ok_or_else(|| {
-            CmsError::UnknownSignatureAlgorithm(signing_cert.signature_algorithm_oid().clone())
-        })?;
-
-        let verification_algorithm =
-            signature_algorithm.resolve_verification_algorithm(key_algorithm)?;
+        let verification_algorithm = self
+            .signature_algorithm
+            .resolve_verification_algorithm(key_algorithm)?;
 
         let public_key = UnparsedPublicKey::new(
             verification_algorithm,
@@ -592,6 +664,55 @@ impl SignerInfo {
         Ok(public_key)
     }
 
+    /// Resolve the time-stamp token [SignedData] for this signer.
+    ///
+    /// The time-stamp token is a SignedData ASN.1 structure embedded as an unsigned
+    /// attribute. This is a convenience method to extract it and turn it into
+    /// a [SignedData].
+    ///
+    /// Returns `Ok(Some)` on success, `Ok(None)` if there is no time-stamp token,
+    /// and `Err` if there is a parsing error.
+    pub fn time_stamp_token_signed_data(&self) -> Result<Option<SignedData>, CmsError> {
+        if let Some(attrs) = self.unsigned_attributes() {
+            if let Some(signed_data) = &attrs.time_stamp_token {
+                Ok(Some(SignedData::try_from(signed_data)?))
+            } else {
+                Ok(None)
+            }
+        } else {
+            Ok(None)
+        }
+    }
+
+    /// Verify the time-stamp token in this instance.
+    ///
+    /// The time-stamp token is a SignedData ASN.1 structure embedded as an unsigned
+    /// attribute. So this method reconstructs that data structure and effectively
+    /// calls [SignerInfo::verify_signature_with_signed_data] and
+    /// [SignerInfo::verify_message_digest_with_signed_data].
+    ///
+    /// Returns `Ok(None)` if there is no time-stamp token and `Ok(Some(()))` if
+    /// there is and the token validates. `Err` occurs on any parse or verification
+    /// error.
+    pub fn verify_time_stamp_token(&self) -> Result<Option<()>, CmsError> {
+        let signed_data = if let Some(v) = self.time_stamp_token_signed_data()? {
+            v
+        } else {
+            return Ok(None);
+        };
+
+        if signed_data.signers.is_empty() {
+            return Ok(None);
+        }
+
+        for signer in signed_data.signers() {
+            signer.verify_signature_with_signed_data(&signed_data)?;
+            signer.verify_message_digest_with_signed_data(&signed_data)?;
+        }
+
+        Ok(Some(()))
+    }
+
     /// Obtain the raw bytes of content that was signed given a `SignedData`.
     ///
     /// This joins the encapsulated content from `SignedData` with `SignedAttributes`
@@ -604,22 +725,52 @@ impl SignerInfo {
     /// Obtain the raw bytes of content that were digested and signed.
     ///
     /// The returned value is the message that was signed and whose signature
-    /// of needs to be verified.
+    /// of which needs to be verified.
     ///
     /// The optional content argument is the `encapContentInfo eContent`
     /// field, typically the value of `SignedData.signed_content()`.
     pub fn signed_content(&self, content: Option<&[u8]>) -> Vec<u8> {
-        let mut res = Vec::new();
+        // Per RFC 5652 Section 5.4:
+        //
+        //    The result of the message digest calculation process depends on
+        //    whether the signedAttrs field is present.  When the field is absent,
+        //    the result is just the message digest of the content as described
+        //    above.  When the field is present, however, the result is the message
+        //    digest of the complete DER encoding of the SignedAttrs value
+        //    contained in the signedAttrs field.  Since the SignedAttrs value,
+        //    when present, must contain the content-type and the message-digest
+        //    attributes, those values are indirectly included in the result.  The
+        //    content-type attribute MUST NOT be included in a countersignature
+        //    unsigned attribute as defined in Section 11.4.  A separate encoding
+        //    of the signedAttrs field is performed for message digest calculation.
+        //    The IMPLICIT [0] tag in the signedAttrs is not used for the DER
+        //    encoding, rather an EXPLICIT SET OF tag is used.  That is, the DER
+        //    encoding of the EXPLICIT SET OF tag, rather than of the IMPLICIT [0]
+        //    tag, MUST be included in the message digest calculation along with
+        //    the length and content octets of the SignedAttributes value.
 
-        if let Some(content) = content {
-            res.extend(content);
+        if let Some(signed_attributes_data) = &self.digested_signed_attributes_data {
+            signed_attributes_data.clone()
+        } else if let Some(content) = content {
+            content.to_vec()
+        } else {
+            vec![]
         }
+    }
 
-        if let Some(signed_data) = &self.digested_signed_attributes_data {
-            res.extend(signed_data.as_slice());
-        }
-
-        res
+    /// Obtain the raw bytes constituting `SignerInfo.signedAttrs` as encoded for signatures.
+    ///
+    /// Cryptographic signatures in the `SignerInfo` ASN.1 type are made from the digest
+    /// of the `EXPLICIT SET OF` DER encoding of `SignerInfo.signedAttrs`, if signed
+    /// attributes are present. This function resolves the raw bytes that are used
+    /// for digest computation and later signing.
+    ///
+    /// This should always be `Some` if the instance was constructed from an ASN.1
+    /// value that had signed attributes.
+    pub fn signed_attributes_data(&self) -> Option<&[u8]> {
+        self.digested_signed_attributes_data
+            .as_ref()
+            .map(|x| x.as_ref())
     }
 
     /// Compute a message digest using a `SignedData` instance.
@@ -673,7 +824,14 @@ impl TryFrom<&crate::asn1::rfc5652::SignerInfo> for SignerInfo {
         };
 
         let digest_algorithm = DigestAlgorithm::try_from(&signer_info.digest_algorithm)?;
-        let signature_algorithm = SignatureAlgorithm::try_from(&signer_info.signature_algorithm)?;
+
+        // The "signature" algorithm can also be a key algorithm identifier. So we
+        // attempt to resolve using the more robust mechanism.
+        let signature_algorithm = SignatureAlgorithm::from_oid_and_digest_algorithm(
+            &signer_info.signature_algorithm.algorithm,
+            digest_algorithm,
+        )?;
+
         let signature = signer_info.signature.to_bytes().to_vec();
 
         let signed_attributes = if let Some(attributes) = &signer_info.signed_attributes {
@@ -791,7 +949,7 @@ impl TryFrom<&crate::asn1::rfc5652::SignerInfo> for SignerInfo {
 /// Represents the contents of a CMS SignedAttributes structure.
 ///
 /// This is a high-level interface to the SignedAttributes ASN.1 type.
-#[derive(Clone, Debug)]
+#[derive(Clone)]
 pub struct SignedAttributes {
     /// The content type of the value being signed.
     ///
@@ -805,6 +963,33 @@ pub struct SignedAttributes {
     signing_time: Option<chrono::DateTime<chrono::Utc>>,
 }
 
+impl SignedAttributes {
+    pub fn content_type(&self) -> &Oid {
+        &self.content_type
+    }
+
+    pub fn message_digest(&self) -> &[u8] {
+        &self.message_digest
+    }
+
+    pub fn signing_time(&self) -> Option<&chrono::DateTime<chrono::Utc>> {
+        self.signing_time.as_ref()
+    }
+}
+
+impl Debug for SignedAttributes {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        let mut s = f.debug_struct("SignedAttributes");
+        s.field("content_type", &format_args!("{}", self.content_type));
+        s.field(
+            "message_digest",
+            &format_args!("{}", hex::encode(&self.message_digest)),
+        );
+        s.field("signing_time", &self.signing_time);
+        s.finish()
+    }
+}
+
 #[derive(Clone, Debug)]
 pub struct UnsignedAttributes {
     /// Time-Stamp Token from a Time-Stamp Protocol server.
@@ -820,7 +1005,9 @@ mod tests {
 
     // This signature was extracted from the Firefox.app/Contents/MacOS/firefox
     // Mach-O executable on a aarch64 machine.
-    const FIREFOX_SIGNATURE: &[u8] = include_bytes!("firefox.der");
+    const FIREFOX_SIGNATURE: &[u8] = include_bytes!("firefox.ber");
+
+    const FIREFOX_CODE_DIRECTORY: &[u8] = include_bytes!("firefox-code-directory");
 
     #[test]
     fn parse_firefox() {
@@ -837,7 +1024,10 @@ mod tests {
 
         let raw2 = crate::asn1::rfc5652::SignedData::decode_ber(&buffer).unwrap();
         assert_eq!(raw, raw2, "BER round tripping is identical");
+    }
 
+    #[test]
+    fn verify_firefox() {
         let signed_data = SignedData::parse_ber(FIREFOX_SIGNATURE).unwrap();
 
         for signer in signed_data.signers.iter() {
@@ -850,6 +1040,23 @@ mod tests {
             signer
                 .verify_message_digest_with_signed_data(&signed_data)
                 .unwrap_err();
+
+            // But we know what that value is. So plug it in to verify.
+            signer
+                .verify_message_digest_with_content(FIREFOX_CODE_DIRECTORY)
+                .unwrap();
+
+            // Now verify the time-stamp token embedded as an unsigned attribute.
+            let tst_signed_data = signer.time_stamp_token_signed_data().unwrap().unwrap();
+
+            for signer in tst_signed_data.signers() {
+                signer
+                    .verify_message_digest_with_signed_data(&tst_signed_data)
+                    .unwrap();
+                signer
+                    .verify_signature_with_signed_data(&tst_signed_data)
+                    .unwrap();
+            }
         }
     }
 }
